#include "network-status-agent.h"

#include <string.h>
#include <NetworkManager.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <iwlib.h>
#include <glibtop/netlist.h>
#include <glibtop/netload.h>

#include "gnome-utils.h"

typedef struct {
	DBusGConnection *nm_conn;
	DBusGProxy      *nm_proxy;
} NetworkStatusAgentPrivate;

#define NETWORK_STATUS_AGENT_GET_PRIVATE(o) (                              \
	G_TYPE_INSTANCE_GET_PRIVATE (                                      \
		(o), NETWORK_STATUS_AGENT_TYPE, NetworkStatusAgentPrivate  \
	)                                                                  \
)

static void network_status_agent_dispose    (GObject *);

static void init_nm_connection (NetworkStatusAgent *);

static NetworkStatusInfo *nm_get_first_active_device_info (NetworkStatusAgent *);
static GList             *nm_get_devices                  (NetworkStatusAgent *);
static NetworkStatusInfo *nm_get_device_info              (NetworkStatusAgent *, DBusGProxy *);

static void              nm_state_change_cb (DBusGProxy *, guint, gpointer);
static DBusHandlerResult nm_message_filter  (DBusConnection *, DBusMessage *, gpointer);

static gboolean string_is_valid_dbus_path (const gchar *);

static NetworkStatusInfo *gtop_get_first_active_device_info (void);

enum {
	STATUS_CHANGED,
	LAST_SIGNAL
};

static guint network_status_agent_signals [LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (NetworkStatusAgent, network_status_agent, G_TYPE_OBJECT)

static void
network_status_agent_class_init (NetworkStatusAgentClass *this_class)
{
	GObjectClass *g_obj_class = (GObjectClass *) this_class;

	g_obj_class->dispose = network_status_agent_dispose;

	network_status_agent_signals [STATUS_CHANGED] = g_signal_new (
		"status-changed",
		G_TYPE_FROM_CLASS (this_class),
		G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
		G_STRUCT_OFFSET (NetworkStatusAgentClass, status_changed),
		NULL,
		NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE,
		0
	);

	g_type_class_add_private (this_class, sizeof (NetworkStatusAgentPrivate));
}

static void
network_status_agent_init (NetworkStatusAgent *agent)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);

	agent->nm_present = FALSE;

	priv->nm_conn  = NULL;
	priv->nm_proxy = NULL;
}

NetworkStatusAgent *
network_status_agent_new ()
{
	return g_object_new (NETWORK_STATUS_AGENT_TYPE, NULL);
}

NetworkStatusInfo *
network_status_agent_get_first_active_device_info (NetworkStatusAgent *agent)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);
	NetworkStatusInfo         *info = NULL;


	if (! priv->nm_conn)
		init_nm_connection (agent);

	if (agent->nm_present)
		info = nm_get_first_active_device_info (agent);
	else
		info = gtop_get_first_active_device_info ();

	return info;
}

static void
network_status_agent_dispose (GObject *obj)
{
	/* FIXME */
}

static void
init_nm_connection (NetworkStatusAgent *agent)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);

	GError *error = NULL;


	priv->nm_conn = dbus_g_bus_get (DBUS_BUS_SYSTEM, & error);

	if (! priv->nm_conn) {
		handle_g_error (& error, "%s: dbus_g_bus_get () failed", __FUNCTION__);

		agent->nm_present = FALSE;

		return;
	}

	if (! dbus_bus_name_has_owner (
		dbus_g_connection_get_connection (priv->nm_conn), NM_DBUS_SERVICE, NULL
	)) {
		agent->nm_present = FALSE;

		return;
	}

	agent->nm_present = TRUE;

	dbus_connection_set_exit_on_disconnect (
		dbus_g_connection_get_connection (priv->nm_conn),
		FALSE
	);

	priv->nm_proxy = dbus_g_proxy_new_for_name (
		priv->nm_conn,
		NM_DBUS_SERVICE, NM_DBUS_PATH, NM_DBUS_INTERFACE
	);

	dbus_g_proxy_add_signal (
		priv->nm_proxy,
		NM_DBUS_SIGNAL_STATE_CHANGE,
		G_TYPE_UINT, G_TYPE_INVALID
	);

	dbus_g_proxy_connect_signal (
		priv->nm_proxy,
		NM_DBUS_SIGNAL_STATE_CHANGE,
		G_CALLBACK (nm_state_change_cb),
		agent,
		NULL
	);

	dbus_connection_add_filter (
		dbus_g_connection_get_connection (priv->nm_conn),
		nm_message_filter, agent, NULL
	);
}

static NetworkStatusInfo *
nm_get_first_active_device_info (NetworkStatusAgent *agent)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);

	NetworkStatusInfo *info = NULL;

	GList *devices;
	GList *node;


	if (! priv->nm_conn)
		return NULL;

	devices = nm_get_devices (agent);

	for (node = devices; node; node = node->next) {
		info = nm_get_device_info (agent, (DBusGProxy *) node->data);

		if (info) {
			if (info->active)
				break;

			g_object_unref (info);

			info = NULL;
		}
	}

	/* FIXME: free contents of devices ? */
	g_list_free (devices);

	return info;
}

static GList *
nm_get_devices (NetworkStatusAgent *agent)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);

	GPtrArray *ptr_array;
	gint       i;

	GList *devices = NULL;

	DBusGProxy *proxy;

	GError *error = NULL;


	dbus_g_proxy_call (
		priv->nm_proxy, "getDevices", & error,
		G_TYPE_INVALID,
		dbus_g_type_get_collection ("GPtrArray", DBUS_TYPE_G_PROXY), & ptr_array,
		G_TYPE_INVALID
	);

	if (error) {
		handle_g_error (& error, "%s: calling \"getDevices\" failed", __FUNCTION__);

		return NULL;
	}

	for (i = 0; i < ptr_array->len; i++) {
		proxy = (DBusGProxy *) g_ptr_array_index (ptr_array, i);

		devices = g_list_append (devices, proxy);
	}
	g_ptr_array_free(ptr_array, TRUE);

	return devices;
}

static NetworkStatusInfo *
nm_get_device_info (NetworkStatusAgent *agent, DBusGProxy *device)
{
	NetworkStatusAgentPrivate *priv = NETWORK_STATUS_AGENT_GET_PRIVATE (agent);
	NetworkStatusInfo         *info = g_object_new (NETWORK_STATUS_INFO_TYPE, NULL);

	DBusGProxy *proxy;
	gchar      *network_path = NULL;

	GError *error = NULL;

	proxy = dbus_g_proxy_new_for_name (
		priv->nm_conn,
		NM_DBUS_SERVICE, dbus_g_proxy_get_path (device), NM_DBUS_INTERFACE_DEVICES
	);

	dbus_g_proxy_call (
		proxy, "getProperties", & error,
		G_TYPE_INVALID,
		DBUS_TYPE_G_PROXY, NULL,
		G_TYPE_STRING,     & info->iface,
		G_TYPE_UINT,       & info->type,
		G_TYPE_STRING,     NULL,
		G_TYPE_BOOLEAN,    & info->active,
		G_TYPE_UINT,       NULL,
		G_TYPE_STRING,     & info->ip4_addr,
		G_TYPE_STRING,     & info->subnet_mask,
		G_TYPE_STRING,     & info->broadcast,
		G_TYPE_STRING,     & info->hw_addr,
		G_TYPE_STRING,     & info->route,
		G_TYPE_STRING,     & info->primary_dns,
		G_TYPE_STRING,     & info->secondary_dns,
		G_TYPE_INT,        NULL,
		G_TYPE_INT,        NULL,
		G_TYPE_BOOLEAN,    NULL,
		G_TYPE_INT,        & info->speed_mbs,
		G_TYPE_UINT,       NULL,
		G_TYPE_UINT,       NULL,
		G_TYPE_STRING,     & network_path,
		G_TYPE_STRV,       NULL,
		G_TYPE_INVALID
	);

	if (error) {
		handle_g_error (& error, "%s: calling \"getProperties\" (A) failed", __FUNCTION__);

		g_free (network_path);
		g_object_unref(info);

		return NULL;
	}

	if (info->active) {
		dbus_g_proxy_call (
			proxy, "getDriver", & error,
			G_TYPE_INVALID,
			G_TYPE_STRING, & info->driver,
			G_TYPE_INVALID
		);

		if (error)
			handle_g_error (& error, "%s: calling \"getDriver\" failed", __FUNCTION__);

		if (info->type == DEVICE_TYPE_802_11_WIRELESS && string_is_valid_dbus_path (network_path)) {
			proxy = dbus_g_proxy_new_for_name (
				priv->nm_conn,
				NM_DBUS_SERVICE, network_path, NM_DBUS_INTERFACE_DEVICES
			);

			dbus_g_proxy_call (
				proxy, "getProperties", & error,
				G_TYPE_INVALID,
				DBUS_TYPE_G_PROXY, NULL,
				G_TYPE_STRING,     & info->essid,
				G_TYPE_STRING,     NULL,
				G_TYPE_INT,        NULL,
				G_TYPE_DOUBLE,     NULL,
				G_TYPE_INT,        NULL,
				G_TYPE_BOOLEAN,    NULL,
				G_TYPE_INT,        NULL,
				G_TYPE_BOOLEAN,    NULL,
				G_TYPE_INVALID
			);
		}

		if (error)
			handle_g_error (& error, "%s: calling \"getProperties\" (B) failed", __FUNCTION__);
	}

	g_free (network_path);

	return info;
}

static void
nm_state_change_cb (DBusGProxy *proxy, guint state, gpointer user_data)
{
	g_signal_emit (
		NETWORK_STATUS_AGENT (user_data),
		network_status_agent_signals [STATUS_CHANGED], 0
	);
}

static DBusHandlerResult
nm_message_filter (DBusConnection *nm_conn, DBusMessage *msg, gpointer user_data)
{
	if (
		(dbus_message_is_signal (msg, DBUS_INTERFACE_LOCAL, "Disconnected") &&
		! strcmp (dbus_message_get_path (msg), DBUS_PATH_LOCAL))            ||
		dbus_message_is_signal (msg, DBUS_INTERFACE_DBUS, "NameOwnerChanged")
	) {
		init_nm_connection (NETWORK_STATUS_AGENT (user_data));

		return DBUS_HANDLER_RESULT_HANDLED;
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/*
 * This macro and function are copied from dbus-marshal-validate.c
 */

#define VALID_NAME_CHARACTER(c) (	\
	((c) >= '0' && (c) <= '9') ||	\
	((c) >= 'A' && (c) <= 'Z') ||	\
	((c) >= 'a' && (c) <= 'z') ||	\
	((c) == '_')			\
)

static gboolean
string_is_valid_dbus_path (const gchar *str)
{
	const gchar *s;
	const gchar *end;
	const gchar *last_slash;

	gint len;


	len = strlen (str);

	s   = str;
	end = s + len;

	if (*s != '/')
		return FALSE;
	last_slash = s;
	++s;

	while (s != end) {
		if (*s == '/') {
			if ((s - last_slash) < 2)
				return FALSE; /* no empty path components allowed */

			last_slash = s;
		}
		else {
			if (! VALID_NAME_CHARACTER (*s))
				return FALSE;
		}
		
		++s;
	}
	
	if ((end - last_slash) < 2 && len > 1)
		return FALSE; /* trailing slash not allowed unless the string is "/" */

	return TRUE;
}

#define CHECK_FLAG(flags, offset) (((flags) & (1 << (offset))) ? TRUE : FALSE)

static NetworkStatusInfo *
gtop_get_first_active_device_info ()
{
	NetworkStatusInfo *info = NULL;

	glibtop_netlist   net_list;
	gchar           **networks;

	glibtop_netload net_load;

	gint            sock_fd;
	wireless_config wl_cfg;

	gint ret;
	gint i;


	networks = glibtop_get_netlist (& net_list);

	for (i = 0; i < net_list.number; ++i) {
		glibtop_get_netload (& net_load, networks [i]);

		if (
			  CHECK_FLAG (net_load.if_flags, GLIBTOP_IF_FLAGS_RUNNING) &&
			! CHECK_FLAG (net_load.if_flags, GLIBTOP_IF_FLAGS_LOOPBACK)
		) {
			sock_fd = iw_sockets_open ();

			if (sock_fd <= 0) {
				g_warning ("error opening socket\n");

				return NULL;
			}

			info = g_object_new (NETWORK_STATUS_INFO_TYPE, NULL);

			ret = iw_get_basic_config (sock_fd, networks [i], & wl_cfg);

			if (ret >= 0) {
				info->type  = DEVICE_TYPE_802_11_WIRELESS;
				info->essid = g_strdup (wl_cfg.essid);
				info->iface = g_strdup (networks [i]);

				return info;
			}
			else {
				info->type  = DEVICE_TYPE_802_3_ETHERNET;
				info->essid = NULL;
				info->iface = g_strdup (networks [i]);

				return info;
			}
		}
	}

	return NULL;
}

#ifdef HHHHHHH
#include "network-status.h"

#include <glib/gi18n.h>
#include <libgnome/gnome-desktop-item.h>

#include "slab-gnome-util.h"
#include "tile-status.h"

#define CHECK_FLAG(flags, offset) (((flags) & (1 << (offset))) ? TRUE : FALSE)

#define CONF_DESKTOP_ITEM_PATH "/usr/share/applications/YaST2/groups/network_devices.desktop"

static GnomeDesktopItem *network_conf_desktop_item = NULL;
void init_network_conf_desktop_item (void);

typedef enum {
	NETWORK_STATUS_WIRED,
	NETWORK_STATUS_WIRELESS
} network_status_type_t;

typedef struct {
	network_status_type_t type;

	gchar *iface_name;
	gchar *essid;
} network_status_t;

network_status_t *get_network_status  (void);
void              network_status_free (network_status_t *status);

void handle_network_conf_open     (GtkWidget   *, gpointer);
void handle_network_conf_help (GtkMenuItem *, gpointer);

network_status_t *get_network_status ()
{
	network_status_t *status;

	glibtop_netlist   net_list;
	gchar           **networks;

	glibtop_netload net_load;

	gint            sock_fd;
	wireless_config wl_cfg;

	gint ret;
	gint i;


	networks = glibtop_get_netlist (& net_list);

	for (i = 0; i < net_list.number; ++i) {
		glibtop_get_netload (& net_load, networks [i]);

		if (
			  CHECK_FLAG (net_load.if_flags, GLIBTOP_IF_FLAGS_RUNNING) &&
			! CHECK_FLAG (net_load.if_flags, GLIBTOP_IF_FLAGS_LOOPBACK)
		) {
			sock_fd = iw_sockets_open ();

			if (sock_fd <= 0) {
				g_warning ("error opening socket\n");

				return NULL;
			}

			status = g_new (network_status_t, 1);

			ret = iw_get_basic_config (sock_fd, networks [i], & wl_cfg);

			if (ret >= 0) {
				status->type       = NETWORK_STATUS_WIRELESS;
				status->essid      = g_strdup (wl_cfg.essid);
				status->iface_name = g_strdup (networks [i]);

				return status;
			}
			else {
				status->type       = NETWORK_STATUS_WIRED;
				status->essid      = NULL;
				status->iface_name = g_strdup (networks [i]);

				return status;
			}
		}
	}

	return NULL;
}

void network_status_free (network_status_t *status)
{
	g_return_if_fail (status != NULL);

	g_free (status->essid);
	g_free (status->iface_name);

	g_free (status);
}

GtkWidget *create_network_status_widget ()
{
	GtkWidget   *tile;
	gchar       *markup;

	network_status_t *status;


	status = get_network_status ();

	tile   = NULL;
	markup = NULL;

	if (status) {
		switch (status->type) {
			case NETWORK_STATUS_WIRED:
				markup = g_strdup_printf (
					_("Using ethernet (%s)"),
					status->iface_name
				);

				tile = tile_status_new (
					"gtk-network",
					GTK_ICON_SIZE_BUTTON,
					_("Network: Wired"),
					markup,
					_("<b>Configure Network</b>"),
					NULL
				);

				g_free (markup);

				break;

			case NETWORK_STATUS_WIRELESS:
				markup = g_strdup_printf (
					_("Connected to: %s"),
					status->essid
				);

				tile = tile_status_new (
					"gtk-network",
					GTK_ICON_SIZE_BUTTON,
					_("Network: Wireless"),
					markup,
					_("<b>Configure Network</b>"),
					NULL
				);

				g_free (markup);

				break;

			default:
				break;
		}
	}

	if (! tile)
		tile = tile_status_new (
			"gtk-network",
			GTK_ICON_SIZE_BUTTON,
			_("Network: None"),
			_("Click to configure the network"),
			_("<b>Configure Network</b>"),
			NULL
		);

	g_signal_connect (
		G_OBJECT (tile),
		"single-click",
		G_CALLBACK (handle_network_conf_open),
		NULL
	);

	g_signal_connect (
		G_OBJECT (tile),
		"context-menu-open-activate",
		G_CALLBACK (handle_network_conf_open),
		NULL
	);

	network_status_free (status);

	return tile;
}

void handle_network_conf_open (GtkWidget *tile, gpointer data)
{
	gboolean  success;


	if (! network_conf_desktop_item)
		init_network_conf_desktop_item ();

	success = open_desktop_item_exec (network_conf_desktop_item);

	if (! success)
		g_warning ("couldn't launch item\n");
}

void handle_network_conf_help (GtkMenuItem *menu_item, gpointer data)
{
	gboolean          success;


	if (! network_conf_desktop_item)
		init_network_conf_desktop_item ();

	success = open_desktop_item_help (network_conf_desktop_item);

	if (! success)
		g_warning ("couldn't open help\n");
}

void init_network_conf_desktop_item ()
{
	GError *error = NULL;

	if (network_conf_desktop_item)
		return;

	network_conf_desktop_item = gnome_desktop_item_new_from_file (
		CONF_DESKTOP_ITEM_PATH,
		0,
		& error
	);

	if (error)
		g_warning (
			"error opening \"%s\" [%s]\n",
			CONF_DESKTOP_ITEM_PATH,
			error->message
		);
}
#endif
